/**
  *  \file game/mine.h
  *  \brief Minefields
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MINE_H
#define PCC_V2_GAME_MINE_H

#include <map>
#include "game/obj.h"
#include "game/objtypes.h"
#include "cpluslib/types.h"
#include "cpluslib/iterator.h"
#include "game/playerset.h"

class GUniverse;
class GMinefieldList;

//! A minefield report.
struct TMinefieldReport {
    int     x, y;                  ///< Center location.
    int     owner;                 ///< Owner.
    bool    web;                   ///< true iff type known to be a web. Otherwise false.
    bool    units_known;           ///< true iff number of units known. false if only radius known.
    bool    type_known;            ///< true iff the type is known reliably. If false, a stored type overrides the new one.
    int32_t radius_or_units;       ///< Number of mine units if %units_known is true, radius if %units_known is false.
    int     turn;                  ///< Turn number of this report.
    int     why;                   ///< Reason for this report (mfa_XXX).
};

//! A Minefield.
class GMinefield : public GCircularObject {
    /* Data of last scan */
    const int id;
    int     x, y;
    int     owner;
    bool    web;
    int32_t units;
    int     turn;

    int     last_action;

    /* Previous data */
    int32_t units_prev;
    int     turn_prev;

    /* Cached current values */
    int     radius_now;
    int32_t units_now;
 public:
    GMinefield(int id);

    void internalCheck(int turnNr);
    void addMinefieldReport(const TMinefieldReport& report);

    /* GMapObject interface */
    string_t getName(Name which) const;
    int      getId() const;
    int      getOwner() const;
    GPoint   getPos() const;
    int      getRadius() const;
    int32_t  getRadiusSquared() const;

    /* Minefield Interface */
    int32_t getUnitsAfterDecay(int32 orig_units) const;
    static int getRadiusFromUnits(int32 units);

    double getPassRate(double distance, bool cloaked, int player) const;

    int getAction() const
        { return last_action; }
    void setAction(int ac)
        { last_action = ac; }
    int32 getUnits() const
        { return units_now; }
    bool isWeb() const
        { return web; }
    int getTurnLastSeen() const
        { return turn; }
    int32_t getUnitsLastSeen() const
        { return units; }
    int32_t getUnitsForLaying() const;

    /** Actions that lead to Minefield Update. These must keep their
        relative and absolute values, because these equal the `ScanType'
        of a UTILx.DAT entry, plus one. */
    enum {
        mfa_None,               ///< No action, i.e. minefield not scanned this turn
        mfa_Laid,               ///< Minefield laid this turn.
        mfa_Swept,              ///< Minefield swept this turn.
        mfa_Scanned             ///< Minefield scanned this turn.
    };
};

/** Minefield Type. Implements the GObjectType interface for minefields,
    and also provides storage for them. */
class GMinefieldType : public GIndexedObjectType {
 public:
    // Types:
    typedef std::map<int,GMinefield*> minemap_t;
    typedef member_iterator<minemap_t::iterator,       GMinefield*,      &minemap_t::value_type::second> ptr_iterator;
    typedef member_iterator<minemap_t::const_iterator, GMinefield*const, &minemap_t::value_type::second> const_ptr_iterator;
    typedef dereferencing_iterator<ptr_iterator>       iterator;
    typedef dereferencing_iterator<const_ptr_iterator> const_iterator;

    // Construction:
    GMinefieldType(GUniverse& univ);
    ~GMinefieldType();

    // GIndexedObjectType:
    GMinefield& getObjectByIndex(int index);
    GUniverse* getUniverseByIndex(int index);
    bool isValidIndex(int index) const;

    // Container interface:
    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;

    void     erase(int i);

    // Load:
    void     addMinefieldReport(int id, const TMinefieldReport& report);
    void     markAllMinefieldsKnown(int player);
    void     internalCheck(int turnNr);

    GMinefield* get(int id) const;

 private:
    void done();

    GUniverse& univ;
    GPlayerSet all_minefields_known;
    minemap_t mines;
};

inline GMinefieldType::iterator
GMinefieldType::begin()
{
    return ptr_iterator(mines.begin());
}

inline GMinefieldType::iterator
GMinefieldType::end()
{
    return ptr_iterator(mines.end());
}

inline GMinefieldType::const_iterator
GMinefieldType::begin() const
{
    return const_ptr_iterator(mines.begin());
}

inline GMinefieldType::const_iterator
GMinefieldType::end() const
{
    return const_ptr_iterator(mines.end());
}

#endif
