/**
  *  \file game/objobserve.h
  *  \brief Class GObjectSelectionObserver
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJOBSERVE_H
#define PCC_V2_GAME_OBJOBSERVE_H

#include "cpluslib/sigslot.h"
#include "game/objsel.h"

class GObject;
class GUniverse;
class GObjectType;

/** Observe a GObjectSelection.
    This class observes a GObjectSelection.
    It provides a virtual onObjectChanged() method which is called
    whenever the object selected into the GObjectSelection changes.
    It also provides methods to access that object. */
class GObjectSelectionObserver {
 public:
    GObjectSelectionObserver(GObjectSelection& selection);
    virtual ~GObjectSelectionObserver();

    GObject* getCurrentObject();
    GUniverse* getCurrentUniverse();
    GObjectSelection& getObjectSelection();
    GObjectType* getObjectType();

    virtual void onObjectChanged() = 0;

 private:
    GObjectSelection& selection;
    SignalConnection conn_current_changed;
    SignalConnection conn_obj_changed;

    void onCurrentChanged();
    void attachObject();
    void attachSelection();
};

/** Get current object.
    \return pointer to object, can be null
    \see GObjectSelection::getCurrentObject() */
inline GObject*
GObjectSelectionObserver::getCurrentObject()
{
    return selection.getCurrentObject();
}

/** Get universe for current object.
    \return pointer to universe, can be null
    \see GObjectSelection::getCurrentUniverse() */
inline GUniverse*
GObjectSelectionObserver::getCurrentUniverse()
{
    return selection.getCurrentUniverse();
}

/** Get object selection.
    Returns a reference to the GObjectSelection this object was constructed with. */
inline GObjectSelection&
GObjectSelectionObserver::getObjectSelection()
{
    return selection;
}

/** Get object type observed by current selection.
    \return pointer to GObjectSelection, can be null */
inline GObjectType*
GObjectSelectionObserver::getObjectType()
{
    return selection.getObjectType();
}

#endif
