/**
  *  \file int/if/plugif.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/plugif.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "util/pluginmgr.h"
#include "util/plugin.h"
#include "int/if/ifutil.h"
#include "int/propenum.h"
#include "cpluslib/format.h"

namespace {
    enum { PluginProperty };
}

static const TIntMapping property_mapping[] = {
    { "DESCRIPTION",        ipiDescription,   PluginProperty, thString },
    { "DIRECTORY",          ipiBaseDirectory, PluginProperty, thString },
    { "ID",                 ipiId,            PluginProperty, thString },
    { "NAME",               ipiName,          PluginProperty, thString },
};

IntPluginContext::IntPluginContext(const string_t& id)
    : id(id)
{ }

IntPluginContext::~IntPluginContext()
{ }

int32_t
IntPluginContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, property_mapping, countof(property_mapping));
}

void
IntPluginContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntPluginContext::get(int32_t index)
{
    return getPluginProperty(id, IntPluginProperty(property_mapping[index].index));
}

bool
IntPluginContext::next()
{
    return false;
}

IntContext*
IntPluginContext::clone()
{
    return new IntPluginContext(id);
}

GObject*
IntPluginContext::getObject()
{
    return 0;
}

void
IntPluginContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(property_mapping, countof(property_mapping));
}

// IntValue:
string_t
IntPluginContext::toString(bool /*readable*/)
{
    return format("#<plugin:%s>", id);
}

void
IntPluginContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/* @q System.Plugin(id:Str):Obj (Function)
   Accesses the properties of the plugin given by the name <tt>Id</tt>.
   If no such plugin was loaded, returns EMPTY.
   @see int:index:group:pluginproperty|Plugin Properties
   @since PCC2 1.99.25 */
IntValue*
IFSystemPlugin(IntArgBlock& args)
{
    args.checkArgumentCount(1);

    string_t id;
    if (!checkStringArg(id, args.getNext())) {
        return 0;
    }

    id = strUCase(id);
    if (!PluginManager::getInstance().getPluginById(id)) {
        return 0;
    }
    return new IntPluginContext(id);
}

IntValue*
getPluginProperty(const string_t& id, IntPluginProperty ipi)
{
    Plugin* p = PluginManager::getInstance().getPluginById(id);
    if (!p) {
        return 0;
    }

    switch (ipi) {
     case ipiId:
        /* @q Id:Str (Plugin Property)
           Id of the plugin.
           @since PCC2 1.99.25 */
        return makeStringValue(p->getId());

     case ipiName:
        /* @q Name:Str (Plugin Property)
           Human-readable name of the plugin.
           @since PCC2 1.99.25 */
        return makeStringValue(p->getName());

     case ipiDescription:
        /* @q Description:Str (Plugin Property)
           Description of the plugin. This can possibly be multiple paragraphs of text.
           @since PCC2 1.99.25 */
        return makeStringValue(p->getDescription());

     case ipiBaseDirectory:
        /* @q Directory:Str (Plugin Property)
           Base directory of the plugin.
           This directory contains files installed with the plugin.
           @since PCC2 1.99.25 */
        return makeStringValue(p->getBaseDirectory());
    }
    return 0;
}
