/**
  *  \file io/io.h
  *  \brief Binary File Parsing
  *
  *  Functions to parse binary files (in particular, those written by
  *  certain BASIC programs).
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FILEIO_IO_H
#define PCC_V2_FILEIO_IO_H

#include <cstddef>
#include "cpluslib/types.h"
#include "cpluslib/buffer.h"
#include "cpluslib/maybe.h"
#include "cpluslib/string.h"
#include "defs.h"               // HAVE_UNALIGNED_OK

class Stream;

/*! \defgroup Maybe_types Maybe types

    Often, we don't know some value. In this case, we store a -1 in
    the data file. To return these, we use the 'mpXX_t' types
    (positive maybe type). Some values can rightfully be -1, though.
    For these, we use 'mnXX_t' (negative maybe type). For convenience,
    we also provide types for char and string which do not have any
    explicitly "forbidden" representation. */
//@{
typedef SimpleMaybe<int16,-1>           mp16_t;     ///< Value which is non-negative or unknown, 16-bit.
typedef SimpleMaybe<int32,-1>           mp32_t;     ///< Value which is non-negative or unknown, 32-bit.
typedef SimpleMaybe<int16,-0x8000>      mn16_t;     ///< Value which is possibly-negative or unknown, 16-bit.
typedef SimpleMaybe<int32,-16*0x8000000>  mn32_t;   ///< Value which is possibly-negative or unknown, 32-bit.
typedef Maybe<char>                     mchar_t;    ///< Character which may be unknown.
typedef Maybe<string_t>                 mstring_t;  ///< String which may be unknown.
//@}

static const int16 UNKNOWNP = -1;                   ///< Unknown value in mp16_t/mp32_t.
static const int16 UNKNOWNS = -0x8000;              ///< Unknown value in mn16_t.
static const char  UNKNOWNC = -1;                   ///< Unknown character value marker.

/** Structure Member Type. */
enum TStructureMember {
    T_END,                      ///< end of definition
    T_INT16,                    ///< signed 16 bit ("WORD")
    T_INT32,                    ///< signed 32 bit ("DWORD")
    T_UINT16,                   ///< unsigned 16 bit
    T_UINT32,                   ///< unsigned 32 bit
    T_STRING,                   ///< string
    T_TYPEREF                   ///< member structure
};

/** Definition of a structure element for automatic packing/unpacking. 
    Defines a single field or array there-of. */
struct TStructureParser {
    TStructureMember type : 16;  ///< Member type.
    std::size_t amount    : 16;  ///< Number of members (more than 1 if array).
    std::size_t offset    : 16;  ///< Offset into C++ structure.
    std::size_t size      : 16;  ///< T_TYPEREF: size of member type (C++)
    const TStructureParser* ref; ///< T_TYPEREF: definition of member type
};

std::size_t getStructureSize(const TStructureParser* parser);
const char* getStructure(const char* src, void* data,
                         const TStructureParser* parser);
const char* getPartialStructure(const char* src, void* data,
                                const TStructureParser* parser,
                                std::size_t offset, std::size_t size);
char* storeStructure(char* dest, const void* data,
                     const TStructureParser* parser);
void storeAppendStructure(Buffer<char>& buf, const void* data,
                          std::size_t size, const TStructureParser* parser);

void storePascalStringT(Stream& s, const string_t& str);
bool getPascalString(Stream& s, string_t& str);
void getPascalStringT(Stream& s, string_t& str);

void storeStructureT(Stream& s, const void* data, const TStructureParser parser[], std::size_t size, char* buffer);
bool getStructure(Stream& s, void* data, const TStructureParser parser[], std::size_t size, char* buffer);
void getStructureT(Stream& s, void* data, const TStructureParser parser[], std::size_t size, char* buffer);


/** Extract structure from character array.
    \param src  [in] character array
    \param data [out] structure
    \return pointer after parsed information */
template<class T>
inline const char*
getStructure(const char* src, T& data)
{
    return getStructure(src, &data, T::parser);
}

/** Extract partial structure from character array. Assumes that we
    read the initial portion of the structure.
    \param src  [in] character array
    \param data [out] structure
    \param size [in] number of valid bytes in \c src
    \return pointer after parsed information, i.e. src+size */
template<class T>
inline const char*
getPartialStructure(const char* src, T& data, std::size_t size)
{
    return getPartialStructure(src, &data, T::parser, 0, size);
}

/** Store structure into character array.
    \param dest [out] character array
    \param data [in] structure
    \return pointer after stored data */
template<class T>
inline char*
storeStructure(char* dest, const T& data)
{
    return storeStructure(dest, &data, T::parser);
}

/** Append structure to character buffer.
    \param buf [in/out] Buffer to append to
    \param data [in] structure */
template<class T>
inline void
storeAppendStructure(Buffer<char>& buf, const T& data)
{
    storeAppendStructure(buf, &data, T::size, T::parser);
}

/** Store structore into stream, throw on error.
    \param s Stream
    \param data [in] structure
    \see storeStructure, Stream::writeT */
template<class T>
inline void
storeStructureT(Stream& s, const T& data)
{
    char buffer[T::size];
    storeStructureT(s, &data, T::parser, T::size, buffer);
}

/** Get structure from stream.
    \param s Stream
    \param data [out] structure
    \return true if success, false if file ends prematurely */
template<class T>
inline bool
getStructure(Stream& s, T& data)
{
    char buffer[T::size];
    return getStructure(s, &data, T::parser, T::size, buffer);
}

/** Get structure from stream.
    \param s Stream
    \param data [out] structure
    \return true if success, false if file ends prematurely */
template<class T>
inline void
getStructureT(Stream& s, T& data)
{
    char buffer[T::size];
    getStructureT(s, &data, T::parser, T::size, buffer);
}

/******************** Word I/O ********************/

/** Parse unsigned 16-bit quantity.
    Parses two bytes, returns unsigned value (0..64k).
    \param src [in] Source, must provide operator[]
    \param dif [in] Index into source array */
template<class T>
inline uint16
getUint16(const T& src, int dif=0)
{
    return (unsigned char) src[dif]
        + 0x100U * (unsigned char) src[dif+1];
}

/** Parse unsigned 32-bit quantity.
    Parses four bytes, returns unsigned value (0..4G).
    \param src [in] Source, must provide operator[]
    \param dif [in] Index into source array */
template<class T>
inline uint32
getUint32(const T& src, int dif=0)
{
    return (unsigned char) src[dif]
        + 0x100U * (unsigned char) src[dif+1]
        + 0x10000U * (unsigned char) src[dif+2]
        + 0x1000000U * (unsigned char) src[dif+3];
}

/** Parse signed 16-bit quantity.
    Parses two bytes, returns signed ("WORD") value (-32k..+32k).
    \param src [in] Source, must provide operator[]
    \param dif [in] Index into source array */
template<class T>
inline int16
getInt16(const T& src, int dif=0)
{
    /* ultra-portable way to parse two's complement -- gcc optimizes
       this away if possible. This may not even be necessary according
       to 4.7p2 [conv.integral]. */
    uint16 u = getUint16(src, dif);
    if(u & 0x8000U)
        return -1-~u;
    else
        return u;
}

/** Parse unsigned 32-bit quantity.
    Parses four bytes, returns signed ("DWORD") value (-2G..+2G)
    \param src [in] Source, must provide operator[]
    \param dif [in] Index into source array */
template<class T>
inline int32
getInt32(const T& src, int dif=0)
{
    uint32 u = getUint32(src, dif);
    if(u & 0x80000000U)
        return -1-~u;
    else
        return u;
}

#ifdef HAVE_UNALIGNED_OK
inline uint16
getUint16(const char* src, int dif=0)
{
    return *(uint16*) (src+dif);
}

inline int16
getInt16(const char* src, int dif=0)
{
    return *(int16*) (src+dif);
}

inline uint32
getUint32(const char* src, int dif=0)
{
    return *(uint32*) (src+dif);
}

inline int32
getInt32(const char* src, int dif=0)
{
    return *(int32*) (src+dif);
}
#endif

/** Store unsigned 16-bit quantity. Stores two bytes.
    \param dest  [out] Destination array
    \param value [in] Value to store. */
inline void
storeUint16(char* dest, uint16 value)
{
#ifdef HAVE_UNALIGNED_OK
    *(uint16*) dest = value;
#else
    unsigned char* d = (unsigned char*) dest;
    *d++ = (value & 255);
    *d   = (value >> 8) & 255;
#endif
}

/** Store unsigned 32-bit quantity. Stores four bytes.
    \param dest  [out] Destination array
    \param value [in] Value to store. */
inline void
storeUint32(char* dest, uint32 value)
{
#ifdef HAVE_UNALIGNED_OK
    *(uint32*) dest = value;
#else
    unsigned char* d = (unsigned char*) dest;
    *d++ = (value & 255);
    *d++ = (value >>= 8) & 255;
    *d++ = (value >>= 8) & 255;
    *d   = (value >> 8) & 255;
#endif
}

/** Store signed 16-bit quantity. Stores two bytes.
    \param dest  [out] Destination array
    \param value [in] Value to store. */
inline void
storeInt16(char* dest, int16 value)
{
    // ISO C++ says int decays into uint according to 2's complement rules
    storeUint16(dest, value);
}

/** Store signed 32-bit quantity. Stores four bytes.
    \param dest  [out] Destination array
    \param value [in] Value to store. */
inline void
storeInt32(char* dest, int32 value)
{
    // ISO C++ says int decays into uint according to 2's complement rules
    storeUint32(dest, value);
}

void storeInt16(char* dest, const int16* src, int count);
void storeUint16(char* dest, const uint16* src, int count);
void storeInt32(char* dest, const int32* src, int count);
void storeUint32(char* dest, const uint32* src, int count);

void getInt16(const char* src, int16* dest, int count);
void getUint16(const char* src, uint16* dest, int count);
void getInt32(const char* src, int32* dest, int count);
void getUint32(const char* src, uint32* dest, int count);

string_t getBasicStringN(const char* name, std::size_t n);
void storeBasicStringN(char* name, std::size_t n, const string_t s);
bool isKnownBasicStringN(char* name, std::size_t n);

void getString(Stream& s, string_t& out, std::size_t n);

#ifdef __BORLANDC__
/* Work around a compiler bug */
#define getBasicString(str) (getBasicStringN((str), sizeof (str)))
#define storeBasicString(str,v) (storeBasicStringN((str), sizeof (str), (v)))
#define isKnownBasicString(str) (isKnownBasicStringN((str), sizeof (str)))
#else
/** Get a BASIC string.
    \param name [in] Array of characters containing the string
    \param N    [in] Array size
    \see getBasicStringN */
template<int N>
inline string_t
getBasicString(const char (&name)[N])
{
    return getBasicStringN(name, N);
}

/** Get a BASIC string. For some reason, g++ can't derive this one
    automatically from the one above.
    \param name [in] Array of characters containing the string
    \param N    [in] Array size
    \see getBasicStringN */
template<int N>
inline string_t
getBasicString(char (&name)[N])
{
    return getBasicStringN(name, N);
}

/** Store BASIC string.
    \param name [out] Array
    \param N    [in] Size of array
    \param s    [in] String to store
    \see storeBasicStringN */
template<int N>
inline void
storeBasicString(char (&name)[N], const string_t s)
{
    storeBasicStringN(name, N, s);
}

/** Check for known BASIC string.
    \param name [in] Array of characters containing the string
    \param N    [in] Array size
    \see isKnownBasicStringN */
template<int N>
inline bool
isKnownBasicString(const char (&name)[N])
{
    return isKnownBasicStringN(name, N);
}

/** Check for known BASIC string.
    \param name [in] Array of characters containing the string
    \param N    [in] Array size
    \see isKnownBasicStringN */
template<int N>
inline bool
isKnownBasicString(char (&name)[N])
{
    return isKnownBasicStringN(name, N);
}
#endif

#endif
