/**
  *  \file u/t_int.cc
  *  \brief Helper routines for Interpreter Testsuite
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_int.h"

#include "int/parseexpr.h"
#include "int/lex.h"
#include "int/exec.h"
#include "int/bytecode.h"
#include "int/data.h"
#include "int/if/ifutil.h"
#include "int/process.h"

int32_t globalVars[3];

/*************************** IntGlobalTestVars ***************************/

int32_t
GlobalTestVars::lookup(const IntNameQuery& name)
{
    if (name.match("A"))
        return 0;
    else if (name.match("B"))
        return 1;
    else if (name.match("C"))
        return 2;
    else
        return -1;
}

IntValue*
GlobalTestVars::get(int32_t index)
{
    return makeIntValue(globalVars[index]);
}

void
GlobalTestVars::set(int32_t index, IntValue* v)
{
    checkIntArg(globalVars[index], v);
}

IntSingleContext*
GlobalTestVars::clone()
{
    return new GlobalTestVars();
}

static void
checkIntegralExpr(const char* expr, int result, bool isBool)
{
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());

        IntExecutionContext exec("checkIntegralExpr");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, ok);

        IntValue* resv = exec.getResult();
        TSM_ASSERT(expr, resv != 0);

        IntIntValue* iv = dynamic_cast<IntIntValue*>(resv);
        TSM_ASSERT(expr, iv != 0);
        TSM_ASSERT_EQUALS(expr, iv->getValue(), result);
        TSM_ASSERT_EQUALS(expr, iv->isBool(), isBool);
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Check expression for integer result.
    \param expr Expression
    \param result Required result */
void
checkIntExpr(const char* expr, int result)
{
    checkIntegralExpr(expr, result, false);
}

/** Check expression for boolean result.
    \param expr Expression
    \param result Required result */
void
checkBoolExpr(const char* expr, int result)
{
    checkIntegralExpr(expr, result, true);
}

/** Check expression for file number result. */
void
checkFileExpr(const char* expr, int result)
{
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());

        IntExecutionContext exec("checkFileExpr");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, ok);

        IntValue* resv = exec.getResult();
        TSM_ASSERT(expr, resv != 0);

        IntFileValue* iv = dynamic_cast<IntFileValue*>(resv);
        TSM_ASSERT(expr, iv != 0);
        TSM_ASSERT_EQUALS(expr, iv->getFileNumber(), result);
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Check expression for null result.
    \param expr Expression */
void
checkNullExpr(const char* expr)
{
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());

        IntExecutionContext exec("checkNullExpr");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, ok);

        TSM_ASSERT(expr, exec.getResult() == 0);
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Check expression for string result.
    \param expr Expression
    \param result Required result */
void
checkStringExpr(const char* expr, const char* result)
{
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());

        IntExecutionContext exec("checkStringExpr");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, ok);

        IntValue* resv = exec.getResult();
        TSM_ASSERT(expr, resv != 0);

        IntStringValue* iv = dynamic_cast<IntStringValue*>(resv);
        TSM_ASSERT(expr, iv != 0);
        TSM_ASSERT_EQUALS(expr, iv->getValue(), result);
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Check expression for float result.
    \param expr Expression
    \param result Required result, within 0.01 of real result */
void
checkFloatExpr(const char* expr, double result)
{
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());

        IntExecutionContext exec("checkFloatExpr");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, ok);

        IntValue* resv = exec.getResult();
        TSM_ASSERT(expr, resv != 0);

        IntFloatValue* iv = dynamic_cast<IntFloatValue*>(resv);
        TSM_ASSERT(expr, iv != 0);
        TSM_ASSERT_LESS_THAN_EQUALS(expr, iv->getValue(), result + 0.01);
        TSM_ASSERT_LESS_THAN_EQUALS(expr, result - 0.01, iv->getValue());
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Check expression for execution error.
    The expression must compile, but may not execute.
    \param expr Expression */
void
checkExecFailure(const char* expr)
{
    bool compiled = false;
    try {
        IntTokenizer tok(expr);
        std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
        TSM_ASSERT_EQUALS(expr, tok.getCurrentToken(), tok.tEnd);

        IntBCORef bco = new IntBytecodeObject();
        node->compileValue(*bco, IntCompilationContext());
        compiled = true;

        IntExecutionContext exec("checkExecFailure");
        exec.pushNewContext(new GlobalTestVars());
        exec.pushNewFrame(new IntExecutionFrame(bco));
        bool ok = runTemporaryProcess(exec);
        TSM_ASSERT(expr, !ok);
    }
    catch (...) {
        TSM_ASSERT(expr, compiled);
    }
}
